import types
import unittest2

from willow.image import Image, setup

from willow.backends.base import ImageBackend


class ImageTestCase(unittest2.TestCase):
    def setUp(self):
        # Reset Image class
        self.reset()

        # Make two fake backends and a bad backend
        class FakeBackend(ImageBackend):
            def to_buffer(self):
                pass

            @classmethod
            def from_buffer(cls, buf):
                return cls()

        class AnotherFakeBackend(ImageBackend):
            def to_buffer(self):
                pass

            @classmethod
            def from_buffer(cls, buf):
                return cls()

        class BadFakeBackend(ImageBackend):
            @classmethod
            def check(cls):
                raise ImportError("Bad image backend")

        self.FakeBackend = FakeBackend
        self.AnotherFakeBackend = AnotherFakeBackend
        self.BadFakeBackend = BadFakeBackend

        Image.register_backend(FakeBackend)
        Image.register_backend(AnotherFakeBackend)
        Image.register_backend(BadFakeBackend)

    @staticmethod
    def tearDownClass():
        # Make sure that Image is returned to its
        # default state after all tests have been run
        ImageTestCase.reset()
        setup(Image)

    @staticmethod
    def reset():
        Image.backends = []
        Image.loaders = {}


class TestRegisterLoader(ImageTestCase):
    """
    Tests the register_loader method on Image
    """
    def test_register_loader(self):
        """
        Tests basic usage of register_loader
        """
        Image.register_loader('.jpg', self.FakeBackend)

        self.assertEqual(Image.loaders, {
            '.jpg': [
                (0, self.FakeBackend),
            ],
        })

    def test_register_loader_priority(self):
        """
        Tests that register_loader saves priority
        """
        Image.register_loader('.jpg', self.FakeBackend, priority=100)

        self.assertEqual(Image.loaders, {
            '.jpg': [
                (100, self.FakeBackend),
            ],
        })

    def test_register_loader_priority_multiple(self):
        """
        Tests that register_loader keeps the loaders list sorted by priority
        """
        Image.register_loader('.jpg', self.FakeBackend, priority=100)
        Image.register_loader('.jpg', self.AnotherFakeBackend, priority=200)

        self.assertEqual(Image.loaders, {
            '.jpg': [
                (100, self.FakeBackend),
                (200, self.AnotherFakeBackend),
            ],
        })

    def test_register_loader_priority_multiple_other_way(self):
        """
        Tests that register_loader keeps the loaders list sorted by priority

        Same as above test, just inserting in opposite way to make sure the
        loaders are being sorted by priority (and not insertion order)
        """
        Image.register_loader('.jpg', self.FakeBackend, priority=200)
        Image.register_loader('.jpg', self.AnotherFakeBackend, priority=100)

        self.assertEqual(Image.loaders, {
            '.jpg': [
                (100, self.AnotherFakeBackend),
                (200, self.FakeBackend),
            ],
        })

    def test_register_loader_different_extension(self):
        """
        Tests that register_loader stores loaders for different extensions
        separately
        """
        Image.register_loader('.jpg', self.FakeBackend)
        Image.register_loader('.gif', self.FakeBackend)

        self.assertEqual(Image.loaders, {
            '.jpg': [
                (0, self.FakeBackend),
            ],
            '.gif': [
                (0, self.FakeBackend),
            ],
        })

    def test_register_loader_different_extension_at_same_time(self):
        """
        Tests that a single backend can be assigned to load two extensions
        with a single call to register_loader
        """
        Image.register_loader(['.jpg', '.gif'], self.FakeBackend)

        self.assertEqual(Image.loaders, {
            '.jpg': [
                (0, self.FakeBackend),
            ],
            '.gif': [
                (0, self.FakeBackend),
            ],
        })

    def test_register_loader_different_extension_at_same_time_tuple(self):
        """
        Tests that a single backend can be assigned to load two extensions with
        a single call to register_loader using a tuple
        """
        Image.register_loader(('.jpg', '.gif'), self.FakeBackend)

        self.assertEqual(Image.loaders, {
            '.jpg': [
                (0, self.FakeBackend),
            ],
            '.gif': [
                (0, self.FakeBackend),
            ],
        })

    def test_register_loader_different_extension_at_same_time_with_priority(self):
        """
        Tests that register_loader saves priority when using multipe extensions
        """
        Image.register_loader(['.jpg', '.gif'], self.FakeBackend, priority=100)

        self.assertEqual(Image.loaders, {
            '.jpg': [
                (100, self.FakeBackend),
            ],
            '.gif': [
                (100, self.FakeBackend),
            ],
        })

    def test_register_loader_priority_same(self):
        """
        Tests that new loaders are placed after loaders that have the same
        priority
        """
        Image.register_loader('.jpg', self.FakeBackend)
        Image.register_loader('.jpg', self.AnotherFakeBackend)

        self.assertEqual(Image.loaders, {
            '.jpg': [
                (0, self.FakeBackend),
                (0, self.AnotherFakeBackend),
            ],
        })

    def test_register_loader_priority_same_other_way(self):
        """
        If multiple backends are inserted with the same priority, the last one
        should be placed last in the list

        Same as above test, just inserting in opposite way to make sure the
        loaders are being sorted by insertion time (and not name)
        """
        Image.register_loader('.jpg', self.AnotherFakeBackend)
        Image.register_loader('.jpg', self.FakeBackend)

        self.assertEqual(Image.loaders, {
            '.jpg': [
                (0, self.AnotherFakeBackend),
                (0, self.FakeBackend),
            ],
        })


class TestFindLoader(ImageTestCase):
    """
    Tests the find_loader method on Image
    """
    def test_find_loader(self):
        """
        Tests basic usage of find_loader
        """
        Image.loaders = {
            '.jpg': [
                (0, self.FakeBackend),
            ],
        }

        self.assertEqual(Image.find_loader('.jpg'), self.FakeBackend)

    def test_find_loader_unknown_extension(self):
        """
        Tests that a LoaderError is raised when find_loader is called with an
        unknown extension
        """
        Image.loaders = {
            '.jpg': [
                (0, self.FakeBackend),
            ],
        }

        self.assertRaises(Image.LoaderError, Image.find_loader, '.jpeg')

    def test_find_loader_multiple_extensions(self):
        """
        Tests that find_loader works well when multiple extensions are
        registered
        """
        Image.loaders = {
            '.jpg': [
                (0, self.FakeBackend),
            ],
            '.gif': [
                (100, self.AnotherFakeBackend),
            ],
        }

        self.assertEqual(Image.find_loader('.jpg'), self.FakeBackend)

    def test_find_loader_priority(self):
        """
        Tests that find_loader gets the backend with the highest priority
        """
        Image.loaders = {
            '.jpg': [
                (100, self.FakeBackend),
                (200, self.AnotherFakeBackend),
            ],
        }

        self.assertEqual(Image.find_loader('.jpg'), self.AnotherFakeBackend)

    def test_find_loader_priority_same(self):
        """
        Tests that find_loader picks the last backend when there are multiple
        backends with the same priority
        """
        Image.loaders = {
            '.jpg': [
                (0, self.FakeBackend),
                (0, self.AnotherFakeBackend),
            ],
        }

        self.assertEqual(Image.find_loader('.jpg'), self.AnotherFakeBackend)

    def test_find_loader_priority_same_other_way(self):
        """
        Tests that find_loader picks the last backend when there are multiple
        backends with the same priority
        """
        Image.loaders = {
            '.jpg': [
                (0, self.AnotherFakeBackend),
                (0, self.FakeBackend),
            ],
        }

        self.assertEqual(Image.find_loader('.jpg'), self.FakeBackend)

    def test_find_loader_with_bad_image_backend(self):
        """
        Tests that find_loader ignores bad backends even if they have a higher
        priority
        """
        Image.loaders = {
            '.jpg': [
                (0, self.FakeBackend),
                (100, self.BadFakeBackend),
            ],
        }

        self.assertEqual(Image.find_loader('.jpg'), self.FakeBackend)

    def test_find_loader_with_only_bad_image_backend(self):
        """
        Tests that find_loader raises a LoaderError when there are no good
        backends available
        """
        Image.loaders = {
            '.jpg': [
                (100, self.BadFakeBackend),
            ],
        }

        self.assertRaises(Image.LoaderError, Image.find_loader, '.jpg')


class TestLoaderDefaultConfiguration(ImageTestCase):
    pass

    # TODO


class TestRegisterOperation(ImageTestCase):
    """
    Tests the register_operation method on Image
    """
    def test_register_operation(self):
        """
        Tests basic usage of register_operation
        """
        @self.FakeBackend.register_operation('test')
        def myop(backend):
            pass

        self.assertEqual(self.FakeBackend.operations, {
            'test': myop,
        })

    def test_register_operation_multiple(self):
        """
        Tests that register_operation can register multiple operations
        """
        @self.FakeBackend.register_operation('test')
        def myop(backend):
            pass

        @self.FakeBackend.register_operation('test2')
        def myotherop(backend):
            pass

        self.assertEqual(self.FakeBackend.operations, {
            'test': myop,
            'test2': myotherop,
        })

    def test_register_operation_multiple_samename(self):
        """
        Tests that register_operation will replace a previously registered
        operation if they have the same name
        """
        @self.FakeBackend.register_operation('test')
        def myop(backend):
            pass

        @self.FakeBackend.register_operation('test')
        def myotherop(backend):
            pass

        self.assertEqual(self.FakeBackend.operations, {
            'test': myotherop,
        })

    def test_register_operation_multiple_samename_other_way(self):
        """
        Tests that register_operation will replace a previously registered
        operation if they have the same name
        """
        @self.FakeBackend.register_operation('test')
        def myotherop(backend):
            pass

        @self.FakeBackend.register_operation('test')
        def myop(backend):
            pass

        self.assertEqual(self.FakeBackend.operations, {
            'test': myop,
        })


class TestFindOperation(ImageTestCase):
    """
    Tests the find_operation method on Image
    """
    def test_find_operation(self):
        """
        Tests basic usage of find_operation
        """
        def myop(backend):
            pass

        self.FakeBackend.operations = {
            'test': myop,
        }

        self.assertEqual(Image.find_operation('test'), (
            self.FakeBackend,
            myop,
        ))

    def test_find_operation_unknown_operation(self):
        """
        Tests find_operation returns None when the operation name is not
        registered
        """
        def myop(backend):
            pass

        self.FakeBackend.operations = {
            'test': myop,
        }

        self.assertIsNone(Image.find_operation('test2'))

    def test_find_operation_multiple_backends(self):
        """
        Tests find_operation picks the first available backend if the operation
        has been registered by multiple backends
        """
        def myop(backend):
            pass

        self.FakeBackend.operations = {
            'test': myop,
        }

        self.AnotherFakeBackend.operations = {
            'test': myop,
        }

        self.assertEqual(Image.find_operation('test'), (
            self.FakeBackend,
            myop,
        ))

    def test_find_operation_multiple_backends_preferred(self):
        """
        Tests find_operation will priorities the preferred backend
        """
        def myop(backend):
            pass

        self.FakeBackend.operations = {
            'test': myop,
        }

        self.AnotherFakeBackend.operations = {
            'test': myop,
        }

        self.assertEqual(Image.find_operation('test', preferred_backend=self.AnotherFakeBackend), (
            self.AnotherFakeBackend,
            myop,
        ))

    def test_find_operation_with_bad_backend(self):
        """
        Tests that find_operation ignores bad backends
        """
        def myop(backend):
            pass

        self.BadFakeBackend.operations = {
            'test': myop,
        }

        self.AnotherFakeBackend.operations = {
            'test': myop,
        }

        self.assertEqual(Image.find_operation('test'), (
            self.AnotherFakeBackend,
            myop,
        ))

    def test_find_operation_with_only_bad_backend(self):
        """
        Tests that find_operation raises a RuntimeError when there are no good
        backends available
        """
        def myop(backend):
            pass

        self.BadFakeBackend.operations = {
            'test': myop,
        }

        self.assertRaises(RuntimeError, Image.find_operation, 'test')


class TestGetAttribute(ImageTestCase):
    """
    This tests the __getattr__ method on Image
    """
    def test_getattr_operation(self):
        """
        Tests that __getattr__ looks up operations correctly
        """
        def myop(backend):
            pass

        self.FakeBackend.operations = {
            'test': myop,
        }

        image = Image(self.FakeBackend())

        self.assertIsInstance(image.test, types.FunctionType)

    def test_getattr_operation_unknown(self):
        """
        Tests that __getattr__ raises an AttributeError when the requested
        attribute is not an operation
        """
        def myop(backend):
            pass

        self.FakeBackend.operations = {
            'test': myop,
        }

        image = Image(self.FakeBackend())

        self.assertRaises(AttributeError, getattr, image, 'test2')


class TestCallOperation(ImageTestCase):
    """
    Tests calling an operation that has been registered in Image
    """
    def test_calls_function(self):
        """
        Tests that calling the operation calls the underlying function
        """
        def myop(backend):
            backend.func_called = True

        self.FakeBackend.operations = {
            'test': myop,
        }

        image = Image(self.FakeBackend())
        image.backend.func_called = False
        image.test()

        self.assertTrue(image.backend.func_called)

    def test_args_get_passed_through(self):
        """
        Tests that args get passed through to the underlying function
        """
        def myop(backend, *args, **kwargs):
            backend.passed_args = args
            backend.passed_kwargs = kwargs

        self.FakeBackend.operations = {
            'test': myop,
        }

        image = Image(self.FakeBackend())
        image.backend.passed_args = None
        image.backend.passed_kwargs = None
        image.test('Hello', 'World', name="Karl")

        self.assertEqual(image.backend.passed_args, ('Hello', 'World'))
        self.assertEqual(image.backend.passed_kwargs, {'name': "Karl"})

    def test_return_value_gets_passed_back(self):
        """
        Tests that the return value of the underlying function gets passed back
        to the caller
        """
        def myop(backend):
            return "Hello world!"

        self.FakeBackend.operations = {
            'test': myop,
        }

        image = Image(self.FakeBackend())

        self.assertEqual(image.test(), "Hello world!")

    def test_switches_backend(self):
        """
        Tests that calling an operation will switch backend if the current
        backend doesn't support it
        """
        def say_hello(backend):
            return "Hello world!"

        self.FakeBackend.operations = {
            'say_hello': say_hello,
        }

        def say_goodbye(backend):
            return "Goodbye!"

        self.AnotherFakeBackend.operations = {
            'say_goodbye': say_goodbye,
        }

        image = Image(self.FakeBackend())
        self.assertIsInstance(image.backend, self.FakeBackend)

        image.say_goodbye()
        self.assertIsInstance(image.backend, self.AnotherFakeBackend)

        image.say_hello()
        self.assertIsInstance(image.backend, self.FakeBackend)


class TestRegisterBackend(ImageTestCase):
    """
    Tests the register_backend method on Image
    """
    def setUp(self):
        super(TestRegisterBackend, self).setUp()
        self.reset()

    def test_register_backend(self):
        """
        Tests basic usage of register_backend
        """
        Image.register_backend(self.FakeBackend)

        self.assertEqual(Image.backends, [self.FakeBackend])

    def test_register_backend_multiple(self):
        """
        Tests register_backend can register multiple backends
        """
        Image.register_backend(self.FakeBackend)
        Image.register_backend(self.AnotherFakeBackend)

        self.assertEqual(Image.backends, [self.FakeBackend, self.AnotherFakeBackend])

    def test_register_backend_removes_duplicates(self):
        """
        Tests register_backend won't insert the same backend more than once
        """
        Image.register_backend(self.FakeBackend)
        Image.register_backend(self.FakeBackend)

        self.assertEqual(Image.backends, [self.FakeBackend])


class TestCheckBackends(ImageTestCase):
    pass


class TestSwitchBackend(ImageTestCase):
    pass

    # TODO


if __name__ == '__main__':
    unittest2.main()



from __future__ import absolute_import

from .base import ImageBackend


class WandBackend(ImageBackend):
    def __init__(self, image):
        self.image = image

    def to_buffer(self):
        return 'RGB', self.image.size, self.image.make_blob('RGB')

# DOESNT WORK. SEE: https://github.com/dahlia/wand/issues/123
#    @classmethod
#    def from_buffer(cls, buf):
#        mode, size, data = buf
#        return cls(Image(blob=data, format=mode, width=size[0], height=size[1]))

    @classmethod
    def from_file(cls, f):
        wand_image = cls.get_wand_image()
        wand_api = cls.get_wand_api()

        f.seek(0)

        image = wand_image.Image(file=f)
        image.wand = wand_api.library.MagickCoalesceImages(image.wand)
        return cls(image)

    @classmethod
    def get_wand_image(cls):
        import wand.image
        return wand.image

    @classmethod
    def get_wand_api(cls):    
        import wand.api
        return wand.api

    @classmethod
    def check(cls):
        cls.get_wand_image()
        cls.get_wand_api()


@WandBackend.register_operation('get_size')
def get_size(backend):
    return backend.image.size


@WandBackend.register_operation('resize')
def resize(backend, width, height):
    backend.image.resize(width, height)


@WandBackend.register_operation('crop')
def crop(backend, left, top, right, bottom):
    backend.image.crop(left=left, top=top, right=right, bottom=bottom)


@WandBackend.register_operation('has_alpha')
def has_alpha(backend):
    return backend.image.alpha_channel


@WandBackend.register_operation('has_animation')
def has_animation(backend):
    return backend.image.animation


@WandBackend.register_operation('get_wand_image')
def get_wand_image(backend):
    return backend.image.clone()



from __future__ import absolute_import

from .base import ImageBackend


class PillowBackend(ImageBackend):
    def __init__(self, image):
        self.image = image

    def to_buffer(self):
        image = self.image

        if image.mode not in ['RGB', 'RGBA']:
            if 'A' in image.mode:
                image = image.convert('RGBA')
            else:
                image = image.convert('RGB')

        return image.mode, image.size, image.tobytes()

    @classmethod
    def from_buffer(cls, buf):
        mode, size, data = buf
        return cls(cls.get_pillow_image().frombytes(mode, size, data))

    def to_file(self, f):
        return self.image.save(f, 'PNG')

    @classmethod
    def from_file(cls, f):
        f.seek(0)
        image = cls.get_pillow_image().open(f)
        image.load()
        return cls(image)

    @classmethod
    def get_pillow_image(cls):
        import PIL.Image
        return PIL.Image

    @classmethod
    def check(cls):
        cls.get_pillow_image()


@PillowBackend.register_operation('get_size')
def get_size(backend):
    return backend.image.size


@PillowBackend.register_operation('resize')
def resize(backend, width, height):
    if backend.image.mode in ['1', 'P']:
        backend.image = backend.image.convert('RGB')

    backend.image = backend.image.resize(
        (width, height), backend.get_pillow_image().ANTIALIAS)


@PillowBackend.register_operation('crop')
def crop(backend, left, top, right, bottom):
    backend.image = backend.image.crop((left, top, right, bottom))


@PillowBackend.register_operation('save_as_jpeg')
def save_as_jpeg(backend, f, quality=85):
    if backend.image.mode in ['1', 'P']:
        backend.image = backend.image.convert('RGB')

    backend.image.save(f, 'JPEG', quality=quality)


@PillowBackend.register_operation('save_as_png')
def save_as_png(backend, f):
    backend.image.save(f, 'PNG')


@PillowBackend.register_operation('has_alpha')
def has_alpha(backend):
    img = backend.image
    return img.mode in ('RGBA', 'LA') or (img.mode == 'P' and 'transparency' in img.info)


@PillowBackend.register_operation('has_animation')
def has_animation(backend):
    # Animation not supported by PIL
    return False


@PillowBackend.register_operation('get_pillow_image')
def get_pillow_image(backend):
    return backend.image.copy()



from __future__ import absolute_import

import os

from .base import ImageBackend


class OpenCVBackend(ImageBackend):
    def __init__(self, image_mode, image_size, image_data):
        self.image_mode = image_mode
        self.image_size = image_size
        self.image_data = image_data

    def to_buffer(self):
        return self.image_mode, self.image_size, self.image_data

    @classmethod
    def from_buffer(cls, buf):
        mode, size, data = buf
        return cls(mode, size, data)

    def opencv_grey_image(self):
        cv = self.get_opencv()

        image = cv.CreateImageHeader(self.image_size, cv.IPL_DEPTH_8U, 3)
        cv.SetData(image, self.image_data)

        grey_image = cv.CreateImage(self.image_size, 8, 1)
        convert_mode = getattr(cv, 'CV_%s2GRAY' % self.image_mode)
        cv.CvtColor(image, grey_image, convert_mode)

        return grey_image

    @classmethod
    def get_opencv(cls):
        try:
            import cv
        except ImportError:
            import cv2.cv as cv

        return cv

    @classmethod
    def check(cls):
        cls.get_opencv()


@OpenCVBackend.register_operation('detect_features')
def detect_features(backend):
    cv = backend.get_opencv()

    image = backend.opencv_grey_image()
    rows = backend.image_size[0]
    cols = backend.image_size[1]

    eig_image = cv.CreateMat(rows, cols, cv.CV_32FC1)
    temp_image = cv.CreateMat(rows, cols, cv.CV_32FC1)
    points = cv.GoodFeaturesToTrack(
        image, eig_image, temp_image, 20, 0.04, 1.0, useHarris=False)

    return points


@OpenCVBackend.register_operation('detect_faces')
def detect_faces(backend):
    cv = backend.get_opencv()

    cascade_filename = os.path.join(
        os.path.dirname(__file__),
        'face_detection',
        'haarcascade_frontalface_alt2.xml',
    )
    cascade = cv.Load(cascade_filename)
    image = backend.opencv_grey_image()

    cv.EqualizeHist(image, image)

    min_size = (40, 40)
    haar_scale = 1.1
    min_neighbors = 3
    haar_flags = 0

    faces = cv.HaarDetectObjects(
        image, cascade, cv.CreateMemStorage(0),
        haar_scale, min_neighbors, haar_flags, min_size
    )

    return [
        (
            face[0][0],
            face[0][1],
            face[0][0] + face[0][2],
            face[0][1] + face[0][3],
        ) for face in faces
    ]



import os.path
from io import BytesIO
import bisect

import six

from willow.backends.pillow import PillowBackend
from willow.backends.wand import WandBackend
from willow.backends.opencv import OpenCVBackend


class Image(object):
    class LoaderError(RuntimeError):
        pass

    def __init__(self, initial_backend):
        self.backend = initial_backend

    def __getattr__(self, attr):
        operation = self.find_operation(attr, preferred_backend=type(self.backend))

        if operation is not None:
            backend, func = operation

            def operation(*args, **kwargs):
                if backend is not type(self.backend):
                    self.switch_backend(backend)

                return func(self.backend, *args, **kwargs)

            return operation
        else:
            raise AttributeError("%r object has no attribute %r" % (
                self.__class__.__name__, attr
            ))

    def switch_backend(self, new_backend):
        if type(self.backend) is new_backend:
            return

        if hasattr(new_backend, 'from_buffer') and hasattr(self.backend, 'to_buffer'):
            buf = self.backend.to_buffer()
            self.backend = new_backend.from_buffer(buf)
            return

        if hasattr(new_backend, 'from_file') and hasattr(self.backend, 'to_file'):
            f = BytesIO()
            self.backend.to_file(f)
            self.backend = new_backend.from_file(f)
            return

        raise RuntimeError("Unable to switch backend")

    @classmethod
    def open(cls, f, initial_backend=None):
        if isinstance(f, six.string_types):
            f = open(f, 'rb')

        if not initial_backend:
            # Work out best initial backend based on file extension
            ext = os.path.splitext(f.name)[1].lower()
            initial_backend = cls.find_loader(ext)

        if initial_backend:
            return cls(initial_backend.from_file(f))

    backends = []

    @classmethod
    def register_backend(cls, backend):
        if backend not in cls.backends:
            cls.backends.append(backend)

    @classmethod
    def check_backends(cls, backends):
        available_backends = []
        unavailable_backends = []

        for backend in backends:
            try:
                backend.check()
            except Exception as e:
                error = "%s: %s" % (type(e).__name__, str(e))
                unavailable_backends.append((backend, error))
            else:
                available_backends.append(backend)

        return available_backends, unavailable_backends

    @classmethod
    def find_operation(cls, operation_name, preferred_backend=None):
        # Try finding the operation in the preferred backend
        if preferred_backend is not None:
            if operation_name in preferred_backend.operations.keys():
                return preferred_backend, preferred_backend.operations[operation_name]

        # Operation doesn't exist in preferred backend, find all backends that implement it
        backends = [backend for backend in cls.backends if operation_name in backend.operations]
        if backends:
            # Now filter that list to only include backends that are available
            available_backends, unavailable_backends = cls.check_backends(backends)

            if available_backends:
                # TODO: Think of a way to select the best backend if multiple backends are found
                # Select the first available backend
                return available_backends[0], available_backends[0].operations[operation_name]

            else:
                # Some backends were found but none of them are available, raise an error
                message = '\n'.join([
                    "The operation '%s' is available in the following backends but they all returned an error:" % operation_name
                ] + [
                    "%s -- %s" % (backend.__name__, error)
                    for backend, error in unavailable_backends
                ])
                raise RuntimeError(message)

    loaders = {}

    @classmethod
    def register_loader(cls, extension, backend, priority=0):
        # If extension is a list or tuple, call this method for each one
        if isinstance(extension, (list, tuple)):
            for ext in extension:
                cls.register_loader(ext, backend, priority=priority)
            return

        # Normalise the extension
        if not extension.startswith('.'):
            extension = '.' + extension

        # Register extension in loaders
        if extension not in cls.loaders:
            cls.loaders[extension] = []

        # Add the backend
        bisect.insort_right(cls.loaders[extension], (priority, backend))

    @classmethod
    def find_loader(cls, extension):
        if extension not in cls.loaders or not cls.loaders[extension]:
            raise cls.LoaderError("Cannot find backend that can load this image file")

        # Find all backends that can load images with this extension
        backends = [backend for priority, backend in cls.loaders[extension]]

        # Now filter that list to only include backends that are available
        available_backends, unavailable_backends = cls.check_backends(backends)

        if available_backends:
            # Select the last available backend
            # The loaders list should be sorted with best backends last
            return available_backends[-1]

        else:
            # Some backends were found but none of them are available, raise an error
            message = '\n'.join([
                "This file can be loaded by the following backends but they all returned an error:"
            ] + [
                "%s -- %s" % (backend.__name__, error)
                for backend, error in unavailable_backends
            ])
            raise cls.LoaderError(message)


def setup(cls):
    # Register builtin image backends
    cls.register_backend(PillowBackend)
    cls.register_backend(WandBackend)
    cls.register_backend(OpenCVBackend)


    # Pillow is very good at loading PNG, JPEG and BMP files
    cls.register_loader(['.png', '.jpg', '.jpeg', '.bmp'], PillowBackend, priority=100)

    # Pillow can load gifs too, but without animation
    cls.register_loader('.gif', PillowBackend, priority=-100)

    # Wand can load PNG, JPEG, BMP and GIF (with animation), but doesn't work as fast as Pillow
    cls.register_loader(['.png', '.jpg', '.jpeg', '.bmp', '.gif'], WandBackend)


setup(Image)



#!/usr/bin/env python

import sys, os

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup


# Hack to prevent "TypeError: 'NoneType' object is not callable" error
# in multiprocessing/util.py _exit_function when setup.py exits
# (see http://www.eby-sarna.com/pipermail/peak/2010-May/003357.html)
try:
    import multiprocessing
except ImportError:
    pass



setup(
    name='Willow',
    version='0.1.2',
    description='A Python image library that sits on top of Pillow, Wand and OpenCV',
    author='Karl Hobley',
    author_email='karlhobley10@gmail.com',
    url='',
    packages=find_packages(),
    include_package_data=True,
    license='BSD',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Multimedia :: Graphics :: Graphics Conversion',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    install_requires=[
        'six>=1.7.0',
    ],
    zip_safe=False,
)



import six


class ImageBackendBase(type):
    def __init__(cls, name, bases, dct):
        super(ImageBackendBase, cls).__init__(name, bases, dct)

        # Make sure all backends have their own operations attribute
        cls.operations = {}

    def __lt__(cls, other):
        # As we insert backend classes into ordered lists, we need to define an
        # ordering for backend classes. This is used incase two priorities are
        # the same leading to the following comparison:
        # (0, BackendA) > (0, BackendB)
        # So this method is simply to prevent a crash in this situation
        return False


class ImageBackend(six.with_metaclass(ImageBackendBase)):
    @classmethod
    def register_operation(cls, operation_name):
        def wrapper(func):
            cls.operations[operation_name] = func

            return func

        return wrapper

    @classmethod
    def check(cls):
        pass



import unittest2

from willow.backends import pillow as pillow_backend


class TestPillowOperations(unittest2.TestCase):
    def setUp(self):
        with open('tests/images/transparent.png', 'rb') as f:
            self.backend = pillow_backend.PillowBackend.from_file(f)

    def test_get_size(self):
        width, height = pillow_backend.get_size(self.backend)
        self.assertEqual(width, 200)
        self.assertEqual(height, 150)

    def test_resize(self):
        pillow_backend.resize(self.backend, 100, 75)
        self.assertEqual(self.backend.image.size, (100, 75))

    def test_crop(self):
        pillow_backend.crop(self.backend, 10, 10, 100, 100)
        self.assertEqual(self.backend.image.size, (90, 90))

    def test_has_alpha(self):
        has_alpha = pillow_backend.has_alpha(self.backend)
        self.assertTrue(has_alpha)

    def test_has_animation(self):
        has_animation = pillow_backend.has_animation(self.backend)
        self.assertFalse(has_animation)
